/*!
 * Copyright (c) <2016> <Andreas Niskanen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "Synthesis.h"

#include <iostream>
#include <fstream>
#include <map>
#include <algorithm>
#include <unistd.h>

using namespace std;

static void show_usage() {
    cout << "USAGE: ./afsynth <file> [options]\n\n"
         << "COMMAND LINE ARGUMENTS:\n"
         << "<file>\t: Input filename for AF synthesis instance in apx format.\n\n"
         << "COMMAND LINE OPTIONS:\n"
         << "-h \t: Display this help message.\n"
         << "-v \t: Display the version of the program.\n"
         << "-o <out>\t: Output clauses to file <out> in wcnf format and exit.\n";
}

static void show_version() {
    cout << "AFSynth (version 2016-08-10)\n"
         << "Synthesizing argumentation frameworks from examples\n";
}

int main(int argc, char **argv)
{
    if (argc == 1) {
        show_usage();
        return 1;
    }

    string filename = argv[1];
    string outfile = "";
    string type = "";
    int flag = 0;

    char tmp;
    while ((tmp = getopt(argc, argv, "f:hm:o:s:t:v")) != -1) {
        switch (tmp) {
            case 'f':
                flag = stoi(optarg);
                break;
            case 'h':
                show_usage();
                return 0;
            case 'v':
                show_version();
                return 0;
            case 'o':
                outfile = optarg;
                break;
            case 't':
                type = optarg;
                break;
        }
    }
/*
    if (outfile != "" && (type != "wcnf" && type != "lp")) {
        cout << "ERROR! Type (wcnf, lp) of output file must be specified.\n";
        show_usage();
        return 1;
    }
*/
    ifstream input;
    input.open(filename);

    if (!input.good()) {
        cout << "ERROR! Cannot open input file.\n";
        return 1;
    }

    Instance inst;
    string line, arg, index, sem, weight;
    map<int,Example> examples;

    while (!input.eof()) {
        getline(input, line);
        line.erase(remove_if(line.begin(), line.end(), ::isspace), line.end());
        if (line.length() == 0 || line[0] == '/' || line[0] == '%') continue;
        if (line.length() < 6) cout << "Warning! Cannot parse line: " << line << "\n";
        string op = line.substr(0,3);
        if (op == "arg") {
            if (line[3] == '(' && line.find(')') != string::npos) {
                arg = line.substr(4,line.find(')')-4);
                inst.add_argument(arg);
            } else {
                cout << "Warning! Cannot parse line: " << line << "\n";
            }
        } else if (op == "sem") {
            if (line[3] == '(' && line.find(',') != string::npos && line.find(')') != string::npos) {
                index = line.substr(4,line.find(',')-4);
                sem = line.substr(line.find(',')+1,line.find(')')-line.find(',')-1);
                examples[stoi(index)].sem = sem;
            } else {
                cout << "Warning! Cannot parse line: " << line << "\n";
            }
        } else if (op == "pos") {
            if (line[3] == '(' && line.find(',') != string::npos && line.find(')') != string::npos) {
                index = line.substr(4,line.find(',')-4);
                arg = line.substr(line.find(',')+1,line.find(')')-line.find(',')-1);
                examples[stoi(index)].add_argument(inst.arg_to_index[arg]);
                examples[stoi(index)].pos = true;
            } else {
                cout << "Warning! Cannot parse line: " << line << "\n";
            }
        } else if (op == "neg") {
            if (line[3] == '(' && line.find(',') != string::npos && line.find(')') != string::npos) {
                index = line.substr(4,line.find(',')-4);
                arg = line.substr(line.find(',')+1,line.find(')')-line.find(',')-1);
                examples[stoi(index)].add_argument(inst.arg_to_index[arg]);
                examples[stoi(index)].pos = false;
            } else {
                cout << "Warning! Cannot parse line: " << line << "\n";
            }
        } else if (op == "wei") {
            if (line[6] == '(' && line.find(',') != string::npos && line.find(')') != string::npos) {
                index = line.substr(7,line.find(',')-7);
                weight = line.substr(line.find(',')+1,line.find(')')-line.find(',')-1);
                examples[stoi(index)].weight = stoi(weight);
            } else {
                cout << "Warning! Cannot parse line: " << line << "\n";
            }
        } else {
            cout << "Warning! Cannot parse line: " << line << "\n";
        }
    }

    for (auto it = examples.begin(); it != examples.end(); ++it) {
        it->second.id = it->first;
        inst.add_example(it->second);
    }

//    inst.print();
    inst.initialize(flag);
    Synthesis::synthesize(inst, outfile, flag);
    if (outfile == "") {
        for (int i = 0; i < inst.examples.size(); i++) {
            if (inst.examples[i].pos) {
                if (inst.examples[i].sat) cout << "Positive example " << inst.examples[i].id << " included.\n";
                else cout << "Positive example " << inst.examples[i].id << " not included.\n";
            } else {
                if (inst.examples[i].sat) cout << "Negative example " << inst.examples[i].id << " included.\n";
                else cout << "Negative example " << inst.examples[i].id << " not included.\n";
            }
        }
        for (int i = 0; i < inst.arguments.size(); i++)
            cout << "arg(" << inst.index_to_arg[inst.arguments[i]] << ").\n";
        for (int i = 0; i < inst.attacks.size(); i++)
            cout << "att(" << inst.index_to_arg[inst.attacks[i].first] << "," << inst.index_to_arg[inst.attacks[i].second] << ").\n";
    }

    return 0;
}
